% This file is part of the Attempto Parsing Engine (APE).
% Copyright 2008, Attempto Group, University of Zurich (see http://attempto.ifi.uzh.ch).
%
% The Attempto Parsing Engine (APE) is free software: you can redistribute it and/or modify it
% under the terms of the GNU Lesser General Public License as published by the Free Software
% Foundation, either version 3 of the License, or (at your option) any later version.
%
% The Attempto Parsing Engine (APE) is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
% PURPOSE. See the GNU Lesser General Public License for more details.
%
% You should have received a copy of the GNU Lesser General Public License along with the Attempto
% Parsing Engine (APE). If not, see http://www.gnu.org/licenses/.


:- module(drs_to_sdrs, [drs_to_sdrs/2]).

/** <module> DRS simplifier

Converts the DRS into a simplified form where every
DRS-box is represented as a list of conditions, i.e.

==
drs(Dom, Conds) --> Conds
==

where Conds is a list of conditions (with sentence IDs).

@author Kaarel Kaljurand
@version 2006-08-15

*/


% Operators used in the DRS.
:- op(400, fx, -).
:- op(400, fx, ~).
:- op(400, fx, <>).
:- op(400, fx, []).
:- op(500, xfx, =>).
:- op(500, xfx, v).


%% drs_to_sdrs(+Drs:drs, -SDrs:sdrs) is det.
%
% Removes the domain-argument from all the (non-atomic) DRS conditions,
% including the embedded ones.
%
% @param Drs is an Attempto DRS
% @param SDrs is an Attempto DRS with the domain-arguments removed

drs_to_sdrs(drs(_, Conds), SConds) :-
	conds_sconds(Conds, SConds).


conds_sconds([], []).

conds_sconds([Cond | Tail], [SCond | STail]) :-
	simplify(Cond, SCond),
	!,
	conds_sconds(Tail, STail).


simplify(drs(_, Conds1) => drs(_, Conds2), =>(SConds1, SConds2)) :-
	conds_sconds(Conds1, SConds1),
	conds_sconds(Conds2, SConds2).

simplify(drs(_, Conds1) v drs(_, Conds2), v(SConds1, SConds2)) :-
	conds_sconds(Conds1, SConds1),
	conds_sconds(Conds2, SConds2).

simplify(-drs(_, Conds), -(SConds)) :-
	conds_sconds(Conds, SConds).

simplify(~drs(_, Conds), ~(SConds)) :-
	conds_sconds(Conds, SConds).

simplify(<>drs(_, Conds), <>(SConds)) :-
	conds_sconds(Conds, SConds).

simplify([]drs(_, Conds), [](SConds)) :-
	conds_sconds(Conds, SConds).

simplify(should(drs(_, Conds)), should(SConds)) :-
	conds_sconds(Conds, SConds).

simplify(Label:drs(_, Conds), Label:SConds) :-
	conds_sconds(Conds, SConds).

simplify(Cond, Cond).
