% This file is part of the Attempto Parsing Engine (APE).
% Copyright 2008, Kaarel Kaljurand <kaljurand@gmail.com>.
%
% The Attempto Parsing Engine (APE) is free software: you can redistribute it and/or modify it
% under the terms of the GNU Lesser General Public License as published by the Free Software
% Foundation, either version 3 of the License, or (at your option) any later version.
%
% The Attempto Parsing Engine (APE) is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
% PURPOSE. See the GNU Lesser General Public License for more details.
%
% You should have received a copy of the GNU Lesser General Public License along with the Attempto
% Parsing Engine (APE). If not, see http://www.gnu.org/licenses/.


:- module(drs_to_dlquery, [
		drs_to_dlquery/2,
		drs_to_dlquery/4,
		drs_to_dlquery/5,
		condlist_to_dlquery/2
	]).

/** <module> DRS to DL-Query translator

Translate an Attempto DRS with a single query/2 condition into
Web Ontology Language (OWL 2) class expression. This class expression
is to be interpreted as DL-Query.

@author Kaarel Kaljurand
@version 2008-09-19

@bug: think about: There is a man. Who does the man like?

*/


:- use_module('../drs_to_sdrs', [
		drs_to_sdrs/2
	]).

:- use_module('../../logger/error_logger', [
		clear_messages/1
	]).

:- use_module(drs_to_owldrs, [
		drs_to_owldrs/2
	]).


:- use_module(drs_to_owlswrl, [
		condlist_and/4,
		get_namedindividual/2
	]).


%:- debug(dvar).
%:- debug(owldrs).


%% drs_to_dlquery(+Drs:term, -DlQuery:term) is det.
%% drs_to_dlquery(+Drs:term, +Name:atom, +Comment:atom, -DlQuery:term) is det.
%% drs_to_dlquery(+Drs:term, +Name:atom, +Comment:atom, -OwlDrs:term, -DlQuery:term) is det.
%
% drs_to_dlquery/[2,4,5] converts an Attempto DRS into DL-Query. In the beginning the DRS is
% modified in order to make the processing more straight-forward. The modified
% DRS (generated by drs_to_owldrs/2) is also returned.
%
% @param Drs is an Attempto DRS
% @param Name is a URI relative to which all entity names are to be interpreted
% @param Comment is a comment to be inserted into the resulting ontology (currently ignored)
% @param OwlDrs is a modified DRS, on the basis of which the DL-Query is generated
% @param DlQuery is a term 'DL-Query'(Name, ClassExpression), where ClassExpression
% is an OWL class expression in OWL FSS (Prolog notation)
%
drs_to_dlquery(Drs, Owl) :-
	drs_to_dlquery(Drs, 'http://attempto.ifi.uzh.ch/ontologies/owlswrl/test', 'DL-Query from ACE text.', _, Owl).

drs_to_dlquery(Drs, Name, Comment, Owl) :-
	drs_to_dlquery(Drs, Name, Comment, _, Owl).

drs_to_dlquery(Drs, Name, _Comment, OwlDrs, 'DL-Query'(Name, ClassExpression)) :-
	copy_term(Drs, DrsCopy),
	clear_messages(owl),
	drs_to_sdrs(DrsCopy, SDrsCopy),
	numbervars(SDrsCopy, 0, _),
	drs_to_owldrs(SDrsCopy, OwlDrs),
	condlist_to_dlquery(OwlDrs, ClassExpression),
	!.

drs_to_dlquery(_, _, _, _, 'DL-Query'('', '')).


%% condlist_to_dlquery(+CondList:term, -ClassExpression:term) is semidet.
%
% Converts a list of DRS conditions into an OWL class expression.
% The DRS must have undergone: drs_to_sdrs, numbervars, drs_to_owldrs.
%
% @param CondList is a list of DRS conditions
% @param ClassExpression is an OWL class expression in OWL FSS (Prolog notation)
%
condlist_to_dlquery(OwlDrs, ClassExpression) :-
	once(select(query(DistinguishedVariable, _)-_, OwlDrs, RemainingOwlDrs)),
	debug(dvar, "Distinguished variable: ~w~n", [DistinguishedVariable]),
	debug(owldrs, "Remaining OWL DRS: ~w~n", [RemainingOwlDrs]),
	preprocess_condlist(RemainingOwlDrs, ToplevelReferents, RemainingOwlDrsWithoutNamed),
	debug(owldrs, "Remaining OWL DRS without named-objects: ~w~n", [RemainingOwlDrsWithoutNamed]),
	condlist_and(DistinguishedVariable, RemainingOwlDrsWithoutNamed, ToplevelReferents, ClassExpression).


%% preprocess_condlist(+CondListIn:list, -RefList:list, -CondListOut:list) is det.
%
% Preprocesses the condition list before it is rolled-up.
%
% 1. Removes named-objects but returns the
% extracted named-objects as a list of ref_oneof/2 terms.
%
% 2. Replaces implications by negations (because rolling-up does not handle implications).
%
% @param CondListIn is a list of DRS conditions
% @param RefList is a list of ref_oneof/2 (containing named-object information)
% @param CondListOut is a list of DRS conditions with named-objects removed
%
% BUG: we currently do not look deeper into the arguments of implications
% although the arguments could also contain implications. Think about that.
%
preprocess_condlist([], [], []).

preprocess_condlist([object(X, Lemma, named, _, _, _)-_ | CondsIn], [ref_oneof(X, 'ObjectOneOf'([Individual])) | RefOneOfTail], CondsOut) :-
	!,
	get_namedindividual(Lemma, Individual),
	preprocess_condlist(CondsIn, RefOneOfTail, CondsOut).

% Implication is turned into negation.
% A => -B == -(A & B)
% Example: What is a country that borders no sea?
preprocess_condlist(['=>'(If, [-Then]) | CondsIn], RefList, [-IfThen | CondsOut]) :-
	!,
	append(If, Then, IfThen),
	preprocess_condlist(CondsIn, RefList, CondsOut).

% Implication is turned into negation.
% A => B == -(A & -B)
% Example: Who owns nothing but cats?
preprocess_condlist(['=>'(If, Then) | CondsIn], RefList, [-[-Then | If] | CondsOut]) :-
	!,
	preprocess_condlist(CondsIn, RefList, CondsOut).

preprocess_condlist([Cond | CondsIn], RefList, [Cond | CondsOut]) :-
	!,
	preprocess_condlist(CondsIn, RefList, CondsOut).
