% This file is part of the Attempto Parsing Engine (APE).
% Copyright 2008, Kaarel Kaljurand <kaljurand@gmail.com>.
%
% The Attempto Parsing Engine (APE) is free software: you can redistribute it and/or modify it
% under the terms of the GNU Lesser General Public License as published by the Free Software
% Foundation, either version 3 of the License, or (at your option) any later version.
%
% The Attempto Parsing Engine (APE) is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
% PURPOSE. See the GNU Lesser General Public License for more details.
%
% You should have received a copy of the GNU Lesser General Public License along with the Attempto
% Parsing Engine (APE). If not, see http://www.gnu.org/licenses/.


:- module(illegal_conditions, [
		illegal_conditions/1,
		illegal_condition/1
	]).

/** <module> DRS checker for DRS-to-OWL/SWRL

Setting some error messages for conditions which ACE->OWL/SWRL cannot handle.

@author Kaarel Kaljurand
@version 2008-09-18

TODO

* add more detailed messages in case of property/[4,5,7]

* paraphrase the DRS to deliver a nicer error message.
Drace needs to support this though (i.e. the situation when the
the DRS contains referents which are defined on the upper level).

* Should error messages be given on the ACE level or on the DRS level?
For the end user the ACE level is better, but on the other hand the input
to ACE->OWL is the DRS (i.e. the module is actually doing DRS->OWL).

*/

:- use_module('../../logger/error_logger.pl', [add_error_message/4]).


% Operators used in the DRS.
:- op(400, fx, -).
:- op(400, fx, ~).
:- op(400, fx, <>).
:- op(400, fx, []).
:- op(500, xfx, v).
:- op(500, xfx, =>).


%% illegal_conditions(+Conditions:list) is det.
%
%
illegal_conditions([]).

illegal_conditions([Cond | CondList]) :-
	illegal_condition(Cond),
	illegal_conditions(CondList).


%% illegal_condition(BoxId, Cond, NewCond) is det.
%
%

illegal_condition(relation(_, of, _)-Id) :-
	add_error_message(owl, Id, of, 'Possessive constructions not supported (in this particular case).').

illegal_condition(modifier_adv(_, Adverb, _)-Id) :-
	add_error_message(owl, Id, Adverb, 'Adverbs not supported.').

illegal_condition(modifier_pp(_, Preposition, _)-Id) :-
	add_error_message(owl, Id, Preposition, 'Prepositional phrases not supported.').

% property/4 (pos/comp/sup adjective)
illegal_condition(property(_, Adjective, _)-Id) :-
	add_error_message(owl, Id, Adjective, 'Adjectives not supported.').

% property/5 (comp_than adjective)
illegal_condition(property(_, Adjective, _, _)-Id) :-
	add_error_message(owl, Id, Adjective, 'Adjectives not supported.').

% property/7 (comp_than + subj/obj adjective)
illegal_condition(property(_, Adjective, _, _, _, _)-Id) :-
	add_error_message(owl, Id, Adjective, 'Adjectives not supported.').

illegal_condition(query(_, QueryWord)-Id) :-
	add_error_message(owl, Id, QueryWord, 'Query not supported.').

illegal_condition(predicate(_, Verb, _)-Id) :-
	add_error_message(owl, Id, Verb, 'Intransitive verbs not supported.').

illegal_condition(predicate(_, Verb, _, _, _)-Id) :-
	add_error_message(owl, Id, Verb, 'Ditransitive verbs not supported.').

illegal_condition([](Drs)) :-
	conds_id(Drs, Id),
	add_error_message(owl, Id, '[]/1', 'Necessity not supported.').

illegal_condition(<>Drs) :-
	conds_id(Drs, Id),
	add_error_message(owl, Id, '<>/1', 'Possibility not supported.').

illegal_condition(_:Drs) :-
	conds_id(Drs, Id),
	add_error_message(owl, Id, ':/2', 'Sentence subordination not supported.').

illegal_condition(~(Drs)) :-
	conds_id(Drs, Id),
	add_error_message(owl, Id, '~/1', 'Negation-as-failure not supported.').

illegal_condition(should(Drs)) :-
	conds_id(Drs, Id),
	add_error_message(owl, Id, 'should/1', 'Recommendation not supported.').

% Note: do not do anything, NP conjunction ('na' as 2nd argument) is handled by the next rule
illegal_condition(has_part(_, _)-_).

illegal_condition(object(_, na, _, _, _, _)-Id) :-
	add_error_message(owl, Id, and, 'Noun phrase conjunctions not supported.').

illegal_condition(object(_, Value, _, _, _, _)-Id) :-
	Value \= na,
	add_error_message(owl, Id, Value, 'A reference to this noun either does not exist or is illegal.').

illegal_condition(predicate(_, Value, _, _)-Id) :-
	add_error_message(owl, Id, Value, 'Subject or object of this verb makes an illegal reference.').


%% conds_id(+Conditions:list, -SentenceId:integer) is det.
%
% Finds the shared sentence ID of the conditions.
%
% BUG: are labeled DRSs not handled here?
%
conds_id([], _).

conds_id([_-Id | _], Id) :- !.

conds_id([Cond | Conds], Id) :-
	functor(Cond, F, 1),
	(F = -; F = <>; F = []; F = ~; F = should),
	!,
	arg(1, Cond, Drs),
	conds_id(Drs, Id),
	conds_id(Conds, Id).

conds_id([Drs1 v Drs2 | Conds], Id) :-
	!,
	conds_id(Drs1, Id),
	conds_id(Drs2, Id),
	conds_id(Conds, Id).

conds_id([Drs1 => Drs2 | Conds], Id) :-
	!,
	conds_id(Drs1, Id),
	conds_id(Drs2, Id),
	conds_id(Conds, Id).
